/*
 * tahitigpio.c
 * platform driver for GPIO in Tahiti platform
 *
 * Copyright (C) 2018 NetApp
 */

#include <linux/init.h>
#include <linux/io.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/mod_devicetable.h>
#include <linux/compat.h>
#include <linux/uaccess.h>
#include <linux/platform_device.h>
#include <linux/dmi.h>
#include <linux/device.h>
#include <linux/kdev_t.h>
#include <linux/pci.h>

#include "tahitigpio.h"
#include "tahitigpio-priv.h"

static platform_gpio_device_t		gpio_device;
static struct class			*gpio_class;
static struct device			*gpio_dev;
static unsigned long			gpio_in_use = 0;
static int				gpio_major;

#define GPIO_COMM1_START	0
#define GPIO_COMM1_END		0x5ff

static int platform_gpio_open(struct inode *inode,
                struct file *file)
{
	if (test_and_set_bit(0, &gpio_in_use)) {
		return -EBUSY;
	}
	return 0;
}

static int platform_gpio_release(struct inode *inode,
                   struct file *file)
{
	clear_bit(0, &gpio_in_use);
	return 0;
}

#define HOST_PAD_OWNERSHIP	0x88
#define PAD_OWNERSHIP		0x38
#define GPIO_NMI_ENABLE		0x178
#define PAD_CONFIGURE		0x588

static long platform_gpio_ioctl(struct file *file,
                unsigned int cmd,
                unsigned long arg)
{
	struct gpioioctl_data	gpio_ioctl_data;
	int			ret = 0;

	switch (cmd) {
	case GPIOIOC_READ:
		if (copy_from_user(&gpio_ioctl_data,
				(void __user *)arg,
				sizeof(gpio_ioctl_data))) {
			return -EFAULT;
		}
		/*
		 * valid register range
		 */
		if (gpio_ioctl_data.offset < GPIO_COMM1_START ||
			gpio_ioctl_data.offset > GPIO_COMM1_END) {
			return -EFAULT;
		}
		gpio_ioctl_data.data = *(volatile uint32_t *)(gpio_device.base_addr +
								gpio_ioctl_data.offset);
		if (copy_to_user((void __user *)arg,
				&gpio_ioctl_data,
				sizeof(gpio_ioctl_data))) {
			ret = -EFAULT;
		}
		break;
	case GPIOIOC_WRITE:
		if (copy_from_user(&gpio_ioctl_data,
                            (void __user *)arg,
                            sizeof(gpio_ioctl_data))) {
			return -EFAULT;
		}
		if (gpio_ioctl_data.offset < GPIO_COMM1_START ||
			gpio_ioctl_data.offset > GPIO_COMM1_END) {
			return -EFAULT;
		}

		/*
		 * Due to the back to back write MMIO cycle restriction.
		 * Thus, all accesses that involve write have to be followed by a read.
		 * refer to doc 549921 LewisBurg BIOS Spec
		 */
		*(volatile uint32_t *)(gpio_device.base_addr +
				   gpio_ioctl_data.offset) =
			gpio_ioctl_data.data;
		gpio_ioctl_data.data = *(volatile uint32_t *)(gpio_device.base_addr +
					gpio_ioctl_data.offset);
	
		break;
	default:
		ret = -EFAULT;
		break;
	}
	return ret;
}	

/*
 * set up GPP-E1 to generate NMI
 */
static void gpio_enable_nmi(void)
{
	uint32_t	data;

	/* Host software pad ownership ACPI mode*/
	data = *(volatile uint32_t *)(gpio_device.base_addr +
					HOST_PAD_OWNERSHIP);
	data &= 0xFFFFFFFD;
	*(volatile uint32_t *)(gpio_device.base_addr +
				HOST_PAD_OWNERSHIP) = data;
	data = *(volatile uint32_t *)(gpio_device.base_addr +
					HOST_PAD_OWNERSHIP);
	/* Pad ownership ACPI mode */
	data = *(volatile uint32_t *)(gpio_device.base_addr +
					PAD_OWNERSHIP);
	data &= 0xFFFFFFFCF;
	*(volatile uint32_t *)(gpio_device.base_addr +
				PAD_OWNERSHIP) = data;
	data = *(volatile uint32_t *)(gpio_device.base_addr +
					PAD_OWNERSHIP);
	/* enable NMI */
	data = *(volatile uint32_t *)(gpio_device.base_addr +
			GPIO_NMI_ENABLE);
	data |= 0x2;
	*(volatile uint32_t *)(gpio_device.base_addr +
			GPIO_NMI_ENABLE) = data;
	data = *(volatile uint32_t *)(gpio_device.base_addr +
			GPIO_NMI_ENABLE);
	/* level, input, Route NMI */
	data = *(volatile uint32_t *)(gpio_device.base_addr +
			PAD_CONFIGURE);
	data &= 0xF9FFF100;
	data |= 0x00020100;
	*(volatile uint32_t *)(gpio_device.base_addr +
			PAD_CONFIGURE) = data;
	data = *(volatile uint32_t *)(gpio_device.base_addr +
			PAD_CONFIGURE);
}

static ssize_t controller_id(struct device *dev,
                        struct device_attribute *attr, char *buf)
{
        uint32_t        data;

        data = *(volatile uint32_t *)(gpio_device.base_addr + 0x10000
                                        + 0x5D0);
        return sprintf(buf, "%d\n", (data >> 1) & 0x1);
}

static DEVICE_ATTR(controller, 0400, controller_id, NULL);

static struct attribute *gpio_default_attrs[] = {
        &dev_attr_controller.attr,
        NULL
};

static struct attribute_group gpio_attr_group = {
        .attrs  = gpio_default_attrs,
        .name   = "tahitigpio",
};

static const struct file_operations platform_gpio_fops = {
        .owner          = THIS_MODULE,
        .unlocked_ioctl = platform_gpio_ioctl,
        .open           = platform_gpio_open,
        .release        = platform_gpio_release,
};

#define INTEL_VID		0x8086
#define P2SB_DID_SUPER		0xA220
#define P2SB_DID_PROD		0xA1A0
#define P2SB_SBURG_HIDE		0xE1
#define P2SB_SBURG_BAR		0x10
#define P2SB_SBURG_BARH		0x14
#define GPIO_COMM1_OFFSET	0xAE0000
#define GPIO_COMM1_SIZE		(64 * 1024)
#define GPIO_COMM0_OFFSET	0xAF0000
#define GPIO_COMM0_SIZE		(64 * 1024)

static int tahiti_gpio_probe(struct platform_device *pdev)
{
	uint32_t		base0;
	uint32_t		base1;
	uint64_t		base64;
	struct resource 	*res;
	void __iomem    	*regbase;
	dev_t			dev;
	int			retval;
	struct pci_dev		*pcidev;

	pcidev = NULL;
	pcidev = pci_get_device(INTEL_VID, P2SB_DID_SUPER, pcidev);
	if (!pcidev) {
		pcidev = pci_get_device(INTEL_VID, P2SB_DID_PROD, pcidev);
	}

	if (!pcidev) {
		return -ENODEV;
	}

	/* Unhide the P2SB device */
	pci_write_config_byte(pcidev, P2SB_SBURG_HIDE, 0);
	pci_read_config_dword(pcidev, P2SB_SBURG_BAR, &base0);
	pci_read_config_dword(pcidev, P2SB_SBURG_BARH, &base1);
	base64 = (((uint64_t)base1) << 32) | (base0 & 0xfffffff0);
	/* Hide the P2Sb device */
	pci_write_config_byte(pcidev, P2SB_SBURG_HIDE, 1);

	gpio_device.base64 = base64;
	base64 |= GPIO_COMM1_OFFSET;
	res = devm_request_mem_region(&pdev->dev, base64,
				GPIO_COMM0_SIZE + GPIO_COMM1_SIZE,
				dev_name(&pdev->dev));
	if (!res) {
		dev_err(&pdev->dev, "Failed to request IO memory\n");
		return -EBUSY;
	}

	regbase = devm_ioremap(&pdev->dev, base64, GPIO_COMM1_SIZE + GPIO_COMM0_SIZE);

	if (!regbase) {
		dev_err(&pdev->dev, "Failed devm_ioremap\n");
		devm_release_mem_region(&pdev->dev, base64,
				GPIO_COMM1_SIZE + GPIO_COMM0_SIZE);
		return -EBUSY;
	}

	gpio_device.base_addr = regbase;
	retval = register_chrdev(0, dev_name(&pdev->dev), &platform_gpio_fops);
	if (retval < 0) {
		dev_err(&pdev->dev, "Failed register_chrdev\n");
		devm_iounmap(&pdev->dev, regbase);
		devm_release_mem_region(&pdev->dev, base64,
					GPIO_COMM1_SIZE + GPIO_COMM0_SIZE);
		return retval;
	}

	gpio_major = retval;
	dev = MKDEV(gpio_major, 0);
	gpio_class = class_create(THIS_MODULE, "gpio");
	gpio_dev = device_create(gpio_class, NULL, dev, NULL, "gpio");
	gpio_enable_nmi();
	sysfs_create_group(&gpio_dev->kobj, &gpio_attr_group);
	return 0;
}

static int tahiti_gpio_remove(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;

	sysfs_remove_group(&dev->kobj, &gpio_attr_group);
	device_destroy(gpio_class, MKDEV(gpio_major, 0));
	class_destroy(gpio_class);
	unregister_chrdev(gpio_major, dev_name(dev));
	devm_iounmap(&pdev->dev, gpio_device.base_addr);
	devm_release_mem_region(&pdev->dev, gpio_device.base64,
				GPIO_COMM1_SIZE + GPIO_COMM0_SIZE);
	return 0;
}	

static struct platform_driver tahiti_gpio_driver = {
	.driver = {
		.name	= "tahitigpio",
		.owner	= THIS_MODULE,
	},
	.probe		= tahiti_gpio_probe,
	.remove		= tahiti_gpio_remove,
};

static struct dmi_system_id platform_dmi_table[] __initdata = {
        {
                .matches = {
                        DMI_MATCH(DMI_SYS_VENDOR,"NetApp"),
                        DMI_MATCH(DMI_PRODUCT_NAME, "Tahiti"),
                }
        },
        {
                .matches = {
                        DMI_MATCH(DMI_SYS_VENDOR,"FOXCONN"),
                        DMI_MATCH(DMI_PRODUCT_NAME, "Tahiti"),
                }
        },
        {}
};
MODULE_DEVICE_TABLE(dmi, platform_dmi_table);

static struct platform_device *mypdev;
static int	__init tahiti_gpio_init(void)
{
	int ret;

	if (!dmi_check_system(platform_dmi_table)) {
		pr_err("tahitigpio: unsupported system.\n");
		return -ENODEV;
	}

	ret = platform_driver_register(&tahiti_gpio_driver);
	if (ret) {
		pr_err("tahitigpio: can't register driver: %d\n", ret);
		return ret;
	}

	mypdev = platform_device_alloc("tahitigpio", -1);
	if (!mypdev) {
		pr_err("tahitigpio: can't allocate device\n");
		platform_driver_unregister(&tahiti_gpio_driver);
		return ENOMEM;
	}

	ret = platform_device_add(mypdev);
	if (ret) {
		pr_err("tahitigpio: can't register device\n");
		platform_device_put(mypdev);
		platform_device_del(mypdev);
		platform_driver_unregister(&tahiti_gpio_driver);
		return ret;
	}

	return 0;
}

static void __exit tahiti_gpio_exit(void)
{
	platform_driver_unregister(&tahiti_gpio_driver);
	platform_device_put(mypdev);
	platform_device_del(mypdev);
}

module_init(tahiti_gpio_init);
module_exit(tahiti_gpio_exit);

MODULE_AUTHOR("Hong Huang <hhuang@netapp.com>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("TAHITI GPIO driver");

